with Strings, Ints, Flt_Vectors;
use Strings, Ints, Flt_Vectors;

pragma Elaborate_All (Strings,Ints,Flt_Vectors);

package body Greens is

   ------------------------------------------------------------------
   --- precision etc ------------------------------------------------
   ------------------------------------------------------------------

   procedure Set_Scalar_Precision(Binary_Digs: in Natural := 0) is
   begin
      if Binary_Digs>0 then
         Scalar_Precision := Binary_Digs;
      end if;
      Reps.Set_Precision(Scalar_Precision);
   end Set_Scalar_Precision;

   procedure Set_Numeric_Precision(Binary_Digs: in Natural := 0) is
   begin
      if Binary_Digs>0 then
         Numeric_Precision := Binary_Digs;
      end if;
      Reps.Set_Precision(Numeric_Precision);
   end Set_Numeric_Precision;

   procedure Set_Precisions(Binary_Digs: in Positive) is
   begin
      Set_Scalar_Precision(Binary_Digs);
      Set_Numeric_Precision(Binary_Digs);
   end Set_Precisions;

   function Effective_Precision(Binary_Digs: Positive) return Positive is
      use Reps;
      NP: constant Positive := Numeric_Precision;
      EP: Positive;
   begin
      Reps.Set_Precision(Binary_Digs);
      EP := Reps.Get_Precision(Reps.Pi); -- involves a computation
      Reps.Set_Precision(NP);
      return EP;
   end Effective_Precision;

   procedure Relax_Checks is
   begin
      Reps.Proper_Rounding(NZero);
      if Verbosity>1 then
         Show0("numeric rounding");
      end if;
   end Relax_Checks;

   procedure Restore_Checks is
   begin
     Proper_Rounding(SZero);
     if Verbosity>1 then
        Show0("scalar rounding");
     end if;
   end Restore_Checks;

   procedure Set_Fourier_Parallel(P: in Boolean := True) is
   begin
     Fourier_Parallel := P;
     SFouriers1.Prod_Parallel := P;
     NFouriers1.Prod_Parallel := P;
   end Set_Fourier_Parallel;

   ------------------------------------------------------------------
   --- conversion of scalar types -----------------------------------
   ------------------------------------------------------------------

   procedure S2N(S: in SMatrix; N: in out NMatrix) is
   begin
      for I in S'Range(1) loop
         for J in S'Range(2) loop S2N(S(I,J),N(I,J)); end loop;
      end loop;
   end S2N;

   procedure N2S(N: in NMatrix; S: in out SMatrix) is
   begin
      for I in N'Range(1) loop
         for J in N'Range(2) loop N2S(N(I,J),S(I,J)); end loop;
      end loop;
   end N2S;

   procedure S2N(S: in SVector; N: in out NVector) is
   begin
      for I in S'Range loop S2N(S(I),N(I)); end loop;
   end S2N;

   procedure N2S(N: in NVector; S: in out SVector) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   procedure N2S(N: in NVector; S: in out SDiag) is
   begin
      for I in N'Range loop N2S(N(I),S(I)); end loop;
   end N2S;

   procedure S2N(S: in SFourier1; N: in out NFourier1) is
   begin
      S2N(S.C,N.C);
      N.R := S.R;
   end S2N;

   procedure N2S(N: in NFourier1; S: in out SFourier1) is
   begin
      N2S(N.C,S.C);
      if not Trunc then SetZero(S.E); end if;
      S.R := N.R;
   end N2S;

   ------------------------------------------------------------------
   --- the real stuff -----------------------------------------------
   ------------------------------------------------------------------

   procedure Show_Param is
   begin
      if Trunc then
         Show0("Scalar type: Numeric");
      else
         Show0("Scalar type: Ball");
      end if;
      Show0("Rep type: " & Info(SZero).RepType);
      Show1("S-Bits =",Effective_Precision(Scalar_Precision));
      Show1("N-Bits =",Effective_Precision(Numeric_Precision));
      Show1("Fourier_Parallel: ",Fourier_Parallel);
      Show1("PDeg =",PDeg);
   end Show_Param;

   procedure SetParam(R1,R2,A: in Flt) is
   begin
      SFouriers1_Green.SetRW(R1,R2,A);
      NFouriers1_Green.SetRW(R1,R2,A);
   end SetParam;
   
   procedure Num_ContrMatrix(P: in SParam; F1,F2: in SFourier1; A: in out SMatrix) is
      use NFouriers1_Green;
      PN: NParam;
      AN: NMatrix(A'Range(1),A'Range(2));
      FN1,FN2: NFourier1;
   begin
      Relax_Checks;
      S2N(P,PN);
      S2N(F1,FN1);
      S2N(F2,FN2);
      ContrMatrix(PN,FN1,FN2,AN);
      N2S(AN,A);
      Restore_Checks;
   end Num_ContrMatrix;

   procedure Num_DerMatrix(P: in SParam; F1,F2: in SFourier1; A: in out SMatrix) is
      use NFouriers1_Green;
      PN: NParam;
      AN: NMatrix(A'Range(1),A'Range(2));
      FN1,FN2: NFourier1;
   begin
      Relax_Checks;
      S2N(P,PN);
      S2N(F1,FN1);
      S2N(F2,FN2);
      DerMatrix(PN,FN1,FN2,AN);
      N2S(AN,A);
      Restore_Checks;
   end Num_DerMatrix;

   procedure Num_EigenVec(A: in SMatrix; Vr,Vi: in out SMatrix) is
     use NSV_RNum;
     Dim: constant Integer := A'Last(1);
     WNr,WNi: NVector(1..Dim);
     AN,VNr,VNi: NMatrix(1..Dim,1..Dim);
   begin
     Relax_Checks;
     S2N(A,AN);
     Eigen(AN,WNr,WNi,VNr,VNi);
     N2S(VNr,Vr);
     N2S(VNi,Vi);
     Restore_Checks;
   end Num_EigenVec;

   procedure Num_EigenVec(A: in SMatrix; Vr: in out SMatrix; Wr: in out SDiag) is
     use NSV, NSV_RNum;
     Dim: constant Integer := A'Last(1);
     WNr,WNi: NVector(1..Dim);
     AN,VNr,VNi: NMatrix(1..Dim,1..Dim);
   begin
     Relax_Checks;
     S2N(A,AN);
     Eigen(AN,WNr,WNi,VNr,VNi);
     N2S(VNr,Vr);
     N2S(WNr,Wr);
     Restore_Checks;
   end Num_EigenVec;

   procedure CheckFix(Dim: in Positive; F1,F2: in out SFourier1; KMax: in Radius) is
     use SFouriers1_Green;
     P: constant SParam := (SZero,SZero);
     A: SMatrix(1..Dim,1..Dim);
   begin
     Num_ContrMatrix(P,F1,F2,A);
     CheckFix(P,A,F1,F2,KMax);
   end CheckFix;
   
   procedure CheckSymm(F: in SFourier1; Div: in Positive) is
      G: SFourier1 := F;
      X1: Flt := Flt(1)/Flt(16);
      X2: Flt := Flt(25)/Flt(8);
      use SFouriers1,SFouriers1_Green,SFouriers1_Green_Quad;
      P: Quadr;
   begin
      if Inf(abs(Val(G,Half*Pi)))<Two**(-10) then
      	 Show0("Greens.CheckSymm error 1.");
	 raise Sorry;
      end if;
      Der(G,Half*G.R);
      P(0) := Scal(X1);
      P(1) := Scal(X2-X1)/Flt(Div);
      P(2) := Scal(0);
      while Inf(P(0))<X2 loop
	 if Inf(Val(Val(G,P)))<Zero then
	    Show0("Greens.CheckSymm error 2.");
	    raise Sorry;
      	 end if;
      	 P(0) := P(0)+Two*P(1);
      end loop;
      SFouriers1.Der(G,Half*G.R);
      X1:= Zero;
      X2:= Flt(1)/Flt(16);
      P(0) := Scal(X1);
      P(1) := Scal(X2-X1)/Flt(Div);
      while Inf(P(0))<X2 loop
	 if Inf(Val(Val(G,P)))<Zero then
	    Show0("Greens.CheckSymm error 3.");
	    raise Sorry;
	 end if;
      	 P(0) := P(0)+Two*P(1);
      end loop;
      X1:= Flt(25)/Flt(8);
      X2:= Flt(26)/Flt(8);
      P(0) := Scal(X1);
      P(1) := Scal(X2-X1)/Flt(Div);
      while Inf(P(0))<X2 loop
	 if Sup(Val(Val(G,P)))>Zero then
	    Show0("Greens.CheckSymm error 4.");
	    raise Sorry;
	 end if;
      	 P(0) := P(0)+Two*P(1);
      end loop;
   end CheckSymm;
   
   procedure CheckEigen(Dim: in Positive; F1,F2: in SFourier1) is
      use SM,SSV;
      P: constant SParam := (SZero,SZero);
      Morse: Integer := 0;
      N: Radius;
      D,Dinv: SDiag(1..Dim);
      A,Vr: SMatrix(1..Dim,1..Dim);
      use SV,SFouriers1,SFouriers1_Green;
      procedure Orthonormalise is new Orthonormalize (ScalProd => ScalProd);
   begin
      Num_DerMatrix(P,F1,F2,A);
      Num_EigenVec(A,Vr,D);
      Orthonormalise(Vr);
      Copy(D,Dinv);
      for K in Dinv'Range loop
      	 Sub(Scal(1),Dinv(K));
      end loop;
      Invert(Dinv);
      N := OpNorm(P,Vr*(D*PsiTrans(Vr)),F1,F2)*OpNorm(F1.R,Vr*(Dinv*PsiTrans(Vr)));     
      if Verbosity>1 then
	 Show1("asymmetry of A",Sup(Norm1(PsiAdjoint(A)-A)));
	 Show1("orthonormal error 1",Sup(COrtho(Vr)));
	 Show1("orthonormal error 2",Sup(Norm1(OrthoErr(Vr))));
	 Show1("A-A",Sup(Norm1(Vr*(D*PsiTrans(Vr))-A)));
	 Show1("norm",N);      
      end if;
      if N<One then
      	 for I in D'Range loop
      	    if Inf(D(I))>One then
      	       Morse := Morse+1;
      	    end if;
      	 end loop;
      	 Show1("The Morse index is ",Morse);
      else
      	 Show0("Index check failed");
      end if;
   end CheckEigen;

   ----------- numerical
   
   procedure Find_Fix(Rho: in Flt; Cas: in Natural) is
      use SV, SFouriers1, SFouriers1_Green;
      Dim: constant Positive := 15;
      Rad: constant Radius := Flt(17)/Flt(16);
      Eps: constant Radius := 1.0e-15;
      A: SMatrix(1..Dim,1..Dim);
      V1,V2: SFourier1;
      P: SParam := (SZero,SZero);
   begin
      if not Trunc then
        raise Sorry with "Green.Find_Fix: run me in numeric mode";
      end if;
      SetRW(Rho);
      if Cas=1 then
      	 SetZero(V1);
      	 V1.C(3) := Scal(Half);
	 V1.C(1) := Scal(1);
	 V1.R := Rad;
      	 P(1) := Scal(1);
      else
      	 Read("in.data",V1);
      end if;
      if Cas=2 then
      	 MapA(P,V1,V2);
      	 V1 := Inv(Sqrt(Norm(V2)))*V1;
      end if;
      MapA(P,V1,V2);
      Show1("norm: ",MaxNorm(V1));
      Show1("diff: ",MaxNorm(V2-V1));
      ApproxFix(P,V1,A,Eps);
      MapA(P,V1,V2);
      Show1("norm: ",MaxNorm(V1));
      Show1("diff: ",MaxNorm(V2-V1));
      Write("out.data",V1);
      Plot(V1);
      Plot3(V1);
      DerMatrix(P,V1,A);
      ShowEigen(A);
   end Find_Fix;

   procedure Find_Fix(Cas: in Natural) is
      use SV,SFouriers1,SFouriers1_Green;
      Dim: constant Positive := 40;  -- must be even
      Rad: constant Radius := Flt(9)/Flt(8);
      Eps: constant Radius := 1.0e-16;
      N: Scalar;
      P: SParam := (SZero,SZero);
      A: SMatrix(1..Dim,1..Dim);
      V1,V2,W1,W2: SFourier1;
   begin
      if not Trunc then
        raise Sorry with "Green.Find_Fix: run me in numeric mode";
      end if;
      if Cas=1 then
      	 SetZero(V1);
      	 SetZero(V2);
      	 V1.C(1) := Scal(1);
	 V2.C(1) := Scal(1);
      	 V1.C(2) := Scal(0.1);
	 V2.C(2) := Scal(0.1);
	 V1.R := Rad;
	 V2.R := Rad;
      	 P(1) := Scal(1);
      else
      	 Read("in.data",V1,V2);
      end if;
      if Cas=2 then
      	 MapA(P,V1,V2,W1,W2);
	 N := Inv(Sqrt(Scal(MaxNorm(W1)+MaxNorm(W2))));
      	 V1 := N*V1;
      	 V2 := N*V2;
      end if;
      MapA(P,V1,V2,W1,W2);
      Show1("norm: ",MaxNorm(V1)+MaxNorm(V2));
      Show1("diff: ",MaxNorm(W1-V1)+MaxNorm(W2-V2));
      ApproxFix(P,V1,V2,A,Eps);
      MapA(P,V1,V2,W1,W2);
      Show1("norm: ",MaxNorm(V1)+MaxNorm(V2));
      Show1("diff: ",MaxNorm(W1-V1)+MaxNorm(W2-V2));
      Write("out.data",V1,V2);
      Show2("norms",MaxNorm(V1),MaxNorm(V2));
      Show2("rhos",GuessRho(V1),GuessRho(V2));
      Plot(V1,V2);
      Plot3(V1,V2);
      DerMatrix(P,V1,V2,A);
      ShowEigen(A);
   end Find_Fix;
   
end Greens;
