with Globals, Ada.Integer_Text_IO;

package body Tasks is
  
  package Txt_IO renames Ada.Text_IO;
  package Int_IO renames Ada.Integer_Text_IO;
  
  procedure Initialize_Lists is
--- assuming MaxLog<100
    S: String(1..2);
  begin
    Postfix(0) := " --- NO";
    for N in Log_Count loop
      Int_IO.Put(S,N);
      if N<10 then S(1) := '0'; end if;
      Postfix(N) := "-log." & S;
      Reserved(N) := False;
    end loop;
  end Initialize_Lists;
  
  function Task_Log(H: in Attribute_Handle) return File_Access is
--- File_Access for Log_File(H.Log)
    Log: Log_Range renames H.Log;
  begin
    if Log /= 0 then
      if Log_IO.Is_Open(Log) then
        return Log_File(Log)'Access;
      end if;
      Txt_IO.Put_Line("Task_Log error:" & Log'Img & " not open");
      Log := 0;
    end if;
    return null;
  end Task_Log;
  
  function Task_Log(ID: in Task_Id := Current_Task) return File_Access is
--- File_Access for Log_File(Reference(ID).Log)
  begin
    return Task_Log(Reference(ID));
  end Task_Log;
  
  function Reserve_Log return Log_Range is
--- mark some unreserved Log_File(N) as reserved and return N
    N: Positive;
  begin
    Log_IO.Reserve(N);
    if N <= MaxLog then return N; end if;
    Txt_IO.Put_Line("Open_log error: too many open files");
    return 0;
  end Reserve_Log;
  
  procedure Free_Log(N: in Log_Range) is
--- mark Log_File(N) as unreserved
  begin
    if N>0 then Log_IO.Free(N); end if;
  end Free_Log;
  
  procedure Proxy_Open_Log(N: in Log_Range; FileName: in String) is
--- open reserved Log_File(N) for another process
    use Globals;
    Err: Boolean;
  begin
    if N=0 or else Filename="" then return; end if;
    Log_IO.Open(N,FileName,Err); -- no verbosity
    if Err then
      Txt_IO.Put_Line("Open_Log error: log" & N'Img & " is already open or was not reserved");
    elsif Verbosity>0 then
      Txt_IO.Put_Line(Log_File(N),"Open_Log info: opened log" & N'Img);
    end if;
  end Proxy_Open_Log;
  
  procedure Open_Log(N: in Log_Range; FileName: in String) is
--- open reserved Log_File(N) and set Reference.Log := N
--- associates log number N with current process
    use Ada.Task_Identification, Globals;
    Err: Boolean;
    Log: Log_Range renames Reference.Log;
  begin
    if N=0 or else Filename="" then return; end if;
    if Log /= 0 then
      if Log_IO.Is_Open(Log) then
        Txt_IO.Put_Line("Open_Log error: old log" & Log'Img & " still open");
      else
        Log := 0;
      end if;
    end if;
    if Log=0 then
      Log_IO.Open(N,FileName,Err); -- no verbosity
      if Err then
        Txt_IO.Put_Line("Open_Log error: log" & N'Img & " is already open or was not reserved");
      else
        Log := N;
        if Verbosity>0 then
          Txt_IO.Put_Line(Log_File(N),"Open_Log info: opened log" & N'Img);
        end if;
      end if;
    end if;
  end Open_Log;
    
  procedure Close_Log(N: in Log_Range; Keep_Reserved: in Boolean := False) is
--- close Log_File(N) and Reinitialize if N=Reference.Log
  begin
    if N /= 0 then
      Log_IO.Close(N,Keep_Reserved); -- includes verbosity
      if N=Reference.Log then
        Reinitialize;
      end if;
    end if;
  end Close_Log;
  
  procedure Close_Log(Keep_Reserved: in Boolean := False) is
  begin
    Close_Log(Reference.Log,Keep_Reserved);
  end Close_Log;
  
  procedure Use_Log(N: in Log_Range) is
--- set Reference.Log := N if Log_File(N) has been reserved and opened
    H: constant Attribute_Handle := Reference;
    Log: Log_Range renames H.Log;
  begin
    if Log /= N then
      if Log=0 then
        if Log_IO.Is_Open(N) then
          Log := N;
        else
          Txt_IO.Put_Line("Use_Log error: log" & N'Img & " is not open");
        end if;
      else
        Txt_IO.Put_Line("Use_Log error: old log" & Log'Img & " still open");
      end if;
    end if;
  end Use_Log;
  
  protected body Protected_Log_Ops is
    
    function Is_Reserved(N: Log_Count) return Boolean is
    begin
      return Reserved(N);
    end Is_Reserved;
    
    procedure Reserve(N: out Positive) is
    begin
      N := 1;
      while Reserved(N) loop
        N := N+1;
        if N>MaxLog then return; end if;
      end loop;
      Reserved(N) := True;
    end Reserve;
    
    procedure Free(N: in Log_Count) is
      F: File_Type renames Log_File(N);
    begin
      if Is_Open(F) then
        Txt_IO.Put_Line("Free error: log" & N'Img & " was not closed");
        Close(F);
      end if;
      Reserved(N) := False;
    end Free;
    
    function Is_Open(N: Log_Count) return Boolean is
    begin
      return Is_Open(Log_File(N));
    end Is_Open;
    
    procedure Open(N: in Log_Count; FileName: in String; Err: out Boolean) is
      F: File_Type renames Log_File(N);
    begin
      if Reserved(N) and then not Is_Open(F) then
        Err := False;
        Open(F,Append_File,FileName & Postfix(N));
      else
        Err := True;
      end if;      
    exception
      when Name_Error => Create(F,Out_File,FileName & Postfix(N));
    end Open;
    
    procedure Close(N: in Log_Count; Keep_Reserved: in Boolean := False) is
      use Globals;
      F: File_Type renames Log_File(N);
    begin
      if Is_Open(F) then
        if Verbosity>0 then
          Txt_IO.Put_Line(F,"Close info: closing log" & N'Img);
        end if;
        Close(F);
        Reserved(N) := Keep_Reserved;
      else
        Txt_IO.Put_Line("Close error: log" & N'Img & " not open");
      end if;
    end Close;
    
  end Protected_Log_Ops;
  
begin
  
  Initialize_Lists;
  
end Tasks;
